@@ -28,12 +28,11 @@
 #![deny(missing_docs)]
 #![warn(noop_method_call)]
 #![deny(unreachable_pub)]
-#![deny(clippy::all)]
+#![warn(clippy::all)]
 #![deny(clippy::await_holding_lock)]
 #![deny(clippy::cargo_common_metadata)]
 #![deny(clippy::cast_lossless)]
 #![deny(clippy::checked_conversions)]
-#![warn(clippy::clone_on_ref_ptr)]
 #![warn(clippy::cognitive_complexity)]
 #![deny(clippy::debug_assert_with_mut_call)]
 #![deny(clippy::exhaustive_enums)]
@@ -57,7 +56,6 @@
 #![deny(clippy::unwrap_used)]
 
 mod err;
-pub mod fallback;
 pub mod params;
 #[cfg(test)]
 mod testing;
@@ -73,9 +71,11 @@ use tor_netdoc::doc::microdesc::{MdDigest, Microdesc};
 use tor_netdoc::doc::netstatus::{self, MdConsensus, RouterStatus};
 use tor_netdoc::types::policy::PortPolicy;
 
+use futures::stream::BoxStream;
 use serde::Deserialize;
-use std::collections::{HashMap, HashSet};
+use std::collections::HashMap;
 use std::net::IpAddr;
+use std::ops::Deref;
 use std::sync::Arc;
 use tracing::warn;
 
@@ -151,59 +151,6 @@ impl SubnetConfig {
     }
 }
 
-/// Internal type: either a microdescriptor, or the digest for a
-/// microdescriptor that we want.
-///
-/// This is a separate type so we can use a HashSet instead of
-/// HashMap.
-#[derive(Clone, Debug)]
-enum MdEntry {
-    /// A microdescriptor that is wanted but not present.
-    Absent {
-        /// Index of the routerstatus entry that wants this microdesc.
-        rs_idx: usize,
-        /// Content digest of the microdescriptor.
-        // TODO: I wanted to make this a reference at first, but that's
-        // nontrival.  At this point I'm not 100% sure it would be a good
-        // idea.
-        d: MdDigest,
-    },
-    /// A microdescriptor that we have.
-    Present {
-        /// The microdescriptor itself.
-        md: Arc<Microdesc>,
-    },
-}
-
-impl std::borrow::Borrow<MdDigest> for MdEntry {
-    fn borrow(&self) -> &MdDigest {
-        self.digest()
-    }
-}
-
-impl MdEntry {
-    /// Return the digest for this entry.
-    fn digest(&self) -> &MdDigest {
-        match self {
-            MdEntry::Absent { d, .. } => d,
-            MdEntry::Present { md, .. } => md.digest(),
-        }
-    }
-}
-
-impl PartialEq for MdEntry {
-    fn eq(&self, rhs: &MdEntry) -> bool {
-        self.digest() == rhs.digest()
-    }
-}
-impl Eq for MdEntry {}
-
-impl std::hash::Hash for MdEntry {
-    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
-        self.digest().hash(state);
-    }
-}
-
 /// An opaque type representing the weight with which a relay or set of
 /// relays will be selected for a given role.
 ///
@@ -275,9 +222,12 @@ pub struct NetDir {
     /// A map from keys to integer values, distributed in the consensus,
     /// and clamped to certain defaults.
     params: NetParameters,
-    /// Map from SHA256 digest of microdescriptors to the
-    /// microdescriptors themselves.
-    mds: HashSet<MdEntry>,
+    /// Map from  routerstatus index (the position of a routerstatus within the
+    /// consensus), to that routerstatus's microdescriptor (if we have one.)
+    mds: Vec<Option<Arc<Microdesc>>>,
+    /// Map from SHA256 of _missing_ microdescriptors to the position of their
+    /// corresponding routerstatus indices within `consensus`.
+    rs_idx_by_missing: HashMap<MdDigest, usize>,
     /// Map from ed25519 identity to index of the routerstatus within
     /// `self.consensus.relays()`.
     ///
@@ -304,6 +254,54 @@ pub struct NetDir {
     weights: weight::WeightSet,
 }
 
+/// An event that a [`NetDirProvider`] can broadcast to indicate that a change in
+/// the status of its directory.
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+#[non_exhaustive]
+pub enum DirEvent {
+    /// A new consensus has been received, and has enough information to be
+    /// used.
+    ///
+    /// This event is also broadcast when a new set of consensus parameters is
+    /// available, even if that set of parameters comes from a configuration
+    /// change rather than from the latest consensus.
+    NewConsensus,
+
+    /// New descriptors have been received for the current consensus.
+    ///
+    /// (This event is _not_ broadcast when receiving new descriptors for a
+    /// consensus which is not yet ready to replace the current consensus.)
+    NewDescriptors,
+}
+
+/// An object that can provide [`NetDir`]s, as well as inform consumers when
+/// they might have changed.
+pub trait NetDirProvider {
+    /// Return a handle to our latest directory, if we have one.
+    fn latest_netdir(&self) -> Option<Arc<NetDir>>;
+
+    /// Return a new asynchronous stream that will receive notification
+    /// whenever the consensus has changed.
+    ///
+    /// Multiple events may be batched up into a single item: each time
+    /// this stream yields an event, all you can assume is that the event has
+    /// occurred at least once.
+    fn events(&self) -> BoxStream<'static, DirEvent>;
+}
+
+impl<T> NetDirProvider for Arc<T>
+where
+    T: NetDirProvider,
+{
+    fn latest_netdir(&self) -> Option<Arc<NetDir>> {
+        self.deref().latest_netdir()
+    }
+
+    fn events(&self) -> BoxStream<'static, DirEvent> {
+        self.deref().events()
+    }
+}
+
 /// A partially build NetDir -- it can't be unwrapped until it has
 /// enough information to build safe paths.
 #[derive(Debug, Clone)]
@@ -343,6 +341,8 @@ pub trait MdReceiver {
     ///
     /// Return true if it was indeed wanted.
     fn add_microdesc(&mut self, md: Microdesc) -> bool;
+    /// Return the number of missing microdescriptors.
+    fn n_missing(&self) -> usize;
 }
 
 impl PartialNetDir {
@@ -374,14 +374,13 @@ impl PartialNetDir {
         // Compute the weights we'll want to use for these relays.
         let weights = weight::WeightSet::from_consensus(&consensus, &params);
 
-        let mds = consensus
+        let n_relays = consensus.relays().len();
+
+        let rs_idx_by_missing = consensus
             .relays()
             .iter()
             .enumerate()
-            .map(|(rs_idx, rs)| MdEntry::Absent {
-                rs_idx,
-                d: *rs.md_digest(),
-            })
+            .map(|(rs_idx, rs)| (*rs.md_digest(), rs_idx))
             .collect();
 
         let rs_idx_by_rsa = consensus
@@ -394,9 +393,10 @@ impl PartialNetDir {
         let netdir = NetDir {
             consensus: Arc::new(consensus),
             params,
-            mds,
+            mds: vec![None; n_relays],
+            rs_idx_by_missing,
             rs_idx_by_rsa: Arc::new(rs_idx_by_rsa),
-            rs_idx_by_ed: HashMap::new(),
+            rs_idx_by_ed: HashMap::with_capacity(n_relays),
             weights,
         };
 
@@ -412,12 +412,9 @@ impl PartialNetDir {
     /// netdir, using the microdescriptors from the previous netdir.
     pub fn fill_from_previous_netdir<'a>(&mut self, prev: &'a NetDir) -> Vec<&'a MdDigest> {
         let mut loaded = Vec::new();
-        for ent in &prev.mds {
-            if let MdEntry::Present { md, .. } = ent {
-                if self.netdir.mds.contains(md.digest()) {
-                    loaded.push(md.digest());
-                    self.netdir.add_arc_microdesc(Arc::clone(md));
-                }
+        for md in prev.mds.iter().flatten() {
+            if self.netdir.add_arc_microdesc(md.clone()) {
+                loaded.push(md.digest());
             }
         }
         loaded
@@ -454,6 +451,9 @@ impl MdReceiver for PartialNetDir {
     fn add_microdesc(&mut self, md: Microdesc) -> bool {
         self.netdir.add_microdesc(md)
     }
+    fn n_missing(&self) -> usize {
+        self.netdir.n_missing()
+    }
 }
 
 impl NetDir {
@@ -467,23 +467,23 @@ impl NetDir {
     /// Return true if we wanted it, and false otherwise.
     #[allow(clippy::missing_panics_doc)] // Can't panic on valid object.
     fn add_arc_microdesc(&mut self, md: Arc<Microdesc>) -> bool {
-        if let Some(prev_ent) = self.mds.take(md.digest()) {
-            if let MdEntry::Absent { rs_idx, .. } = prev_ent {
-                assert_eq!(self.consensus.relays()[rs_idx].md_digest(), md.digest());
+        if let Some(rs_idx) = self.rs_idx_by_missing.remove(md.digest()) {
+            assert_eq!(self.consensus.relays()[rs_idx].md_digest(), md.digest());
 
-                // There should never be two approved MDs in the same
-                // consensus listing the same ID... but if there is,
-                // we'll let the most recent one win.
-                self.rs_idx_by_ed.insert(*md.ed25519_id(), rs_idx);
+            // There should never be two approved MDs in the same
+            // consensus listing the same ID... but if there is,
+            // we'll let the most recent one win.
+            self.rs_idx_by_ed.insert(*md.ed25519_id(), rs_idx);
 
-                // Happy path: we did indeed want this one.
-                self.mds.insert(MdEntry::Present { md });
+            // Happy path: we did indeed want this one.
+            self.mds[rs_idx] = Some(md);
 
-                return true;
-            } else {
-                // We already had this.
-                self.mds.insert(prev_ent);
+            // Save some space in the missing-descriptor list.
+            if self.rs_idx_by_missing.len() < self.rs_idx_by_missing.capacity() / 4 {
+                self.rs_idx_by_missing.shrink_to_fit();
             }
+
+            return true;
         }
 
         // Either we already had it, or we never wanted it at all.
@@ -491,15 +491,21 @@ impl NetDir {
     }
 
     /// Construct a (possibly invalid) Relay object from a routerstatus and its
-    /// microdescriptor (if any).
-    fn relay_from_rs<'a>(
+    /// position within the consensus.
+    fn relay_from_rs_and_idx<'a>(
         &'a self,
         rs: &'a netstatus::MdConsensusRouterStatus,
+        rs_idx: usize,
     ) -> UncheckedRelay<'a> {
-        let md = match self.mds.get(rs.md_digest()) {
-            Some(MdEntry::Present { md }) => Some(Arc::as_ref(md)),
-            _ => None,
-        };
+        debug_assert_eq!(
+            self.consensus.relays()[rs_idx].rsa_identity(),
+            rs.rsa_identity()
+        );
+        let md = self.mds[rs_idx].as_deref();
+        if let Some(md) = md {
+            debug_assert_eq!(rs.md_digest(), md.digest());
+        }
+
         UncheckedRelay { rs, md }
     }
 
@@ -527,7 +533,8 @@ impl NetDir {
         self.consensus
             .relays()
             .iter()
-            .map(move |rs| self.relay_from_rs(rs))
+            .enumerate()
+            .map(move |(idx, rs)| self.relay_from_rs_and_idx(rs, idx))
     }
     /// Return an iterator over all usable Relays.
     pub fn relays(&self) -> impl Iterator<Item = Relay<'_>> {
@@ -542,10 +549,10 @@ impl NetDir {
     /// with this ID, the ID may become usable.
     #[allow(clippy::missing_panics_doc)] // Can't panic on valid object.
     pub fn by_id(&self, id: &Ed25519Identity) -> Option<Relay<'_>> {
-        let rs_idx = self.rs_idx_by_ed.get(id)?;
-        let rs = self.consensus.relays().get(*rs_idx).expect("Corrupt index");
+        let rs_idx = *self.rs_idx_by_ed.get(id)?;
+        let rs = self.consensus.relays().get(rs_idx).expect("Corrupt index");
 
-        let relay = self.relay_from_rs(rs).into_relay()?;
+        let relay = self.relay_from_rs_and_idx(rs, rs_idx).into_relay()?;
         assert_eq!(id, relay.id());
         Some(relay)
     }
@@ -611,10 +618,10 @@ impl NetDir {
     /// Return a (possibly unusable) relay with a given RSA identity.
     #[allow(clippy::missing_panics_doc)] // Can't panic on valid object.
     pub fn by_rsa_id_unchecked(&self, rsa_id: &RsaIdentity) -> Option<UncheckedRelay<'_>> {
-        let rs_idx = self.rs_idx_by_rsa.get(rsa_id)?;
-        let rs = self.consensus.relays().get(*rs_idx).expect("Corrupt index");
+        let rs_idx = *self.rs_idx_by_rsa.get(rsa_id)?;
+        let rs = self.consensus.relays().get(rs_idx).expect("Corrupt index");
         assert_eq!(rs.rsa_identity(), rsa_id);
-        Some(self.relay_from_rs(rs))
+        Some(self.relay_from_rs_and_idx(rs, rs_idx))
     }
     /// Return the relay with a given RSA identity, if we have one
     /// and it is usable.
@@ -865,17 +872,14 @@ impl NetDir {
 
 impl MdReceiver for NetDir {
     fn missing_microdescs(&self) -> Box<dyn Iterator<Item = &MdDigest> + '_> {
-        Box::new(self.consensus.relays().iter().filter_map(move |rs| {
-            let d = rs.md_digest();
-            match self.mds.get(d) {
-                Some(MdEntry::Absent { d, .. }) => Some(d),
-                _ => None,
-            }
-        }))
+        Box::new(self.rs_idx_by_missing.keys())
     }
     fn add_microdesc(&mut self, md: Microdesc) -> bool {
         self.add_arc_microdesc(Arc::new(md))
     }
+    fn n_missing(&self) -> usize {
+        self.rs_idx_by_missing.len()
+    }
 }
 
 impl<'a> UncheckedRelay<'a> {
